from .parsers import *
from .util import (
    add_braces_if_no_braces,
    dot2int,
    int2dot,
    seq2int,
    standardize_sequence,
    sequence_has_regular_characters,
)
import numpy as np
from .rnastructure import RNAstructure_singleton
from .env import env


class Datapoint:

    """A datapoint is a data structure where:
    - reference is the name of the sequence in the fasta file
    - sequence is a string of A, C, G, U
    - structure is a string of (, ), .
    - dms is a list of floats corresponding to the reactivity of each base in the sequence
    - paired_bases is a list of tuples (i,j) where i and j are paired bases.


    Example:
    >>> print(Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0]))
    "reference":{"sequence": "AACCGG", "paired_bases": [[0, 5], [1, 4]], "dms": [1.0, 2.0, 3.0]}
    >>> print(Datapoint(reference='reference', sequence='not a valid sequence', structure='((..))'))
    None
    >>> print(Datapoint(reference='reference', sequence='NNNNNNN', structure='((..))'))
    None
    >>> print(Datapoint(reference='reference', sequence='AACCGG', paired_bases=[(0, 5), (1, 4)]))
    "reference":{"sequence": "AACCGG", "paired_bases": [[0, 5], [1, 4]]}
    >>> print(Datapoint(reference='reference', sequence='AACCGG', paired_bases=[(1, 5), (1, 4)]))
    None
    """

    def __new__(cls, *args, **kwargs):
        # Check if the conditions are met before creating the object
        sequence = kwargs.get("sequence", args[0] if len(args) > 0 else None)
        reference = kwargs.get("reference", args[1] if len(args) > 1 else None)
        if sequence is None or reference is None:
            return None
        sequence = standardize_sequence(sequence)

        if (
            not sequence_has_regular_characters(sequence)
            or len(sequence) == 0
            or len(reference) == 0
        ):
            return None

        if "paired_bases" in kwargs:
            if not cls._assert_paired_bases(None, kwargs["paired_bases"], sequence):
                return None

        # If conditions are met, create and return the new instance
        instance = super().__new__(cls)
        return instance

    def __init__(
        self, sequence, reference, structure=None, dms=None, paired_bases=None
    ):
        for attr in [sequence, reference]:
            assert isinstance(
                attr, str
            ), f"Expected {attr} to be a string, got {type(attr)} instead."
            assert len(attr) > 0, f"Expected {attr} to be non-empty."

        # standardize the sequence
        sequence = standardize_sequence(sequence)

        if not sequence_has_regular_characters(sequence):
            raise Exception(
                f"Sequence {sequence} contains characters other than ACGTUacgtu."
            )

        self.reference = reference
        self.sequence = sequence
        self.structure = structure
        if paired_bases is not None or structure is not None:
            self.paired_bases = (
                self._format_paired_bases(paired_bases)
                if paired_bases is not None
                else self._format_paired_bases(
                    self.structure_to_paired_bases(structure)
                )
            )
        if dms is not None:
            self.dms = self._format_dms(dms)
        self.opt_dict = {
            attr: eval(f"self.{attr}")
            for attr in ["paired_bases", "dms"]
            if hasattr(self, attr)
        }

    def _format_paired_bases(self, paired_bases):
        """Returns a set of tuples.
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> datapoint._format_paired_bases([[1, 4], [0, 5]])
        ((0, 5), (1, 4))
        """
        return (
            tuple(sorted([tuple(sorted(pair)) for pair in paired_bases]))
            if paired_bases is not None
            else None
        )

    def _assert_paired_bases(self, paired_bases, sequence):
        """Ensures that the base pairs are 0=<bp<len(sequence) and that that there's maximum one pair per base.
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> datapoint._assert_paired_bases([(0, 5), (1, 4)], 'AACCGG')
        True
        >>> datapoint._assert_paired_bases([(0, 5), (1, 4), (1, 2)], 'AACCGG')
        False
        >>> datapoint._assert_paired_bases([(0, 5), (4, 1), (1, 2)], 'AACCGG')
        False
        >>> datapoint._assert_paired_bases([(0, 8), (3, 4), (1, 2)], 'AACCGG')
        False
        >>> datapoint._assert_paired_bases(None, 'AACCGG')
        True
        """
        if paired_bases is not None:
            return (
                all(
                    [
                        0 <= pair[0] < len(sequence)
                        and 0 <= pair[1] < len(sequence)
                        and pair[0] != pair[1]
                        for pair in paired_bases
                    ]
                )
                and len(
                    set(
                        [pair[0] for pair in paired_bases]
                        + [pair[1] for pair in paired_bases]
                    )
                )
                == len(paired_bases) * 2
            )
        else:
            return True

    def _format_dms(self, dms):
        """returns a tuple
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> datapoint._format_dms([1.0, 2.0, 3.0])
        (1.0, 2.0, 3.0)
        """
        if type(dms) == np.ndarray:
            dms = dms.tolist()
        return tuple(dms)
    
    def embed_dms(self):
        """Returns a list of floats corresponding to the dms.

        >>> from numpy import array, float32
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> assert (datapoint.embed_dms() == array([1., 2., 3.], dtype=float32)).all(), 'The dms are not embedded correctly.'
        """
        return np.array([round(d,4) for d in self.dms], dtype=np.float32)

    def to_dict(self):
        return self.reference, {
            "sequence": self.sequence,
            **{k: v for k, v in self.opt_dict.items() if v is not None},
        }

    def to_flat_dict(self):
        return {
            "reference": self.reference,
            "sequence": self.sequence,
            **{k: v for k, v in self.opt_dict.items() if v is not None},
        }

    @classmethod
    def from_flat_dict(cls, d):
        return cls(
            sequence=d["sequence"],
            reference=d["reference"],
            **{k: v for k, v in d.items() if k not in ["sequence", "reference"]},
        )

    def __str__(self):
        return (
            '"'
            + self.reference
            + '"'
            + ":"
            + str(
                {
                    "sequence": self.sequence,
                    **{k: v for k, v in self.opt_dict.items() if v is not None},
                }
            )
            .replace("'", '"')
            .replace("(", "[")
            .replace(")", "]")
            .replace("None", "null")
        )

    def __repr__(self) -> str:
        out = (
            f"{self.__class__.__name__}('{self.reference}', sequence='{self.sequence}'"
        )
        if hasattr(self, "paired_bases"):
            out += f", paired_bases={self.paired_bases}"
        if hasattr(self, "dms"):
            out += f", dms={self.dms}"
        return out + ")"

    def structure_to_paired_bases(self, structure):
        """Returns a list of tuples (i,j) where i and j are paired bases."""
        paired_bases = set()
        stack = []
        for i, char in enumerate(structure):
            if char == "(":
                stack.append(i)
            elif char == ")":
                paired_bases.add((stack.pop(), i))
        return paired_bases

    def embed_sequence(self):
        """Returns a list of integers corresponding to the sequence.

        >>> from numpy import array
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> assert (datapoint.embed_sequence() == array([1, 1, 2, 2, 3, 3])).all(), 'The sequence is not embedded correctly.'
        """
        return np.array([seq2int[base] for base in self.sequence], dtype=np.int8)

    def embed_structure(self):
        """Returns a list of integers corresponding to the structure.

        >>> from numpy import array
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> assert not (datapoint.embed_structure() - array([2, 2, 1, 1, 3, 3])).any(), 'The sequence is not embedded correctly.'
        """
        return np.array([dot2int[base] for base in self.structure], dtype=np.int8)


class DatapointFactory:
    """A datapoint is a tuple (sequence, reference, structure, dms) where:

    - reference is the name of the sequence in the fasta file
    - sequence is a string of A, C, G, U
    - structure is a string of (, ), .
    - dms is a list of floats corresponding to the reactivity of each base in the sequence

    """

    def from_ct(ct_file, predict_dms):
        """Create a datapoint from a ct file. If predict_dms is True, the dms will be predicted using RNAstructure"""
        reference, sequence, paired_bases = Ct.parse(ct_file)
        sequence = standardize_sequence(sequence)

        if sequence_has_regular_characters(sequence):
            return Datapoint(
                sequence,
                reference,
                paired_bases=paired_bases,
                dms=Ct.predict_dms(ct_file) if predict_dms else None,
            )

    def from_fasta(sequence, reference, predict_structure, predict_dms):
        """Create a datapoint from a fasta file. The structure and dms will be None."""
        sequence = standardize_sequence(sequence)

        if sequence_has_regular_characters(sequence):
            structure, dms = None, None
            if predict_structure:
                structure = Fasta.predict_structure(sequence)
            if predict_dms:
                dms = Fasta.predict_dms(sequence)
            return Datapoint(sequence, reference, structure, dms)

    def from_dreem_output(reference, sequence, mutation_rate, predict_structure):
        """Create a datapoint from a dreem output file. The structure and dms will be predicted if predict_structure and predict_dms are True."""
        sequence = standardize_sequence(sequence)
        mutation_rate = np.array([float(m) for m in mutation_rate], dtype=np.float32)
        if sequence_has_regular_characters(sequence):
            return Datapoint(
                sequence=sequence,
                reference=reference,
                structure=RNAstructure_singleton.predictStructure(
                    sequence, dms=mutation_rate
                )
                if predict_structure
                else None,
                dms=mutation_rate,
            )

    def from_json_line(ref, d, predict_structure=False, predict_dms=False):
        """Create a datapoint from a json line. The json line should have the following format:
        "reference": {"sequence": "sequence", "paired_bases": [[1, 2], [3,4]], "dms": [1.0, 2.0, 3.0]}

        Example:
        >>> DatapointFactory.from_json_line("reference", {"sequence": "ACAAGU"})
        Datapoint('reference', sequence='ACAAGU')
        >>> DatapointFactory.from_json_line("reference", {"sequence": "ACAAGU", "paired_bases": [[1, 2], [3, 4]], "dms": [1.0, 2.0, 3.0]})
        Datapoint('reference', sequence='ACAAGU', paired_bases=((1, 2), (3, 4)), dms=(1.0, 2.0, 3.0))
        >>> DatapointFactory.from_json_line("reference", {"sequence": "something else than ACGTUacgtu", "paired_bases": [[1, 2], [3, 4]], "dms": [1.0, 2.0, 3.0]})
        """

        # create the datapoint
        sequence = d["sequence"]
        sequence = standardize_sequence(sequence)

        if predict_structure and (not "structure" in d) and (not "paired_bases" in d):
            d["structure"] = RNAstructure_singleton.predictStructure(sequence)

        if predict_dms and (not "dms" in d):
            d["dms"] = RNAstructure_singleton.predictPairingProbability(sequence)

        if sequence_has_regular_characters(sequence):
            return Datapoint(
                reference=ref,
                sequence=sequence,
                structure=d["structure"] if "structure" in d else None,
                paired_bases=d["paired_bases"] if "paired_bases" in d else None,
                dms=d["dms"] if "dms" in d else None,
            )




import os
import numpy as np
from .datapoint import Datapoint, DatapointFactory
from typing import List
from .parsers import Fasta, DreemOutput
from .util import UKN

import pandas as pd
from sklearn.metrics import roc_auc_score
from tqdm import tqdm as tqdm_parser
import json

class ListofDatapoints:

    def __init__(self, datapoints=[], verbose=True):
        self.datapoints, self.filtering_report = self.filter_duplicates_and_unvalid(datapoints, verbose)
        
    def __call__(self) -> List[Datapoint]:
        return self.datapoints

    @classmethod
    def from_fasta(cls, fasta_file, predict_structure, predict_dms, tqdm=True, verbose=True):
        """Create a list of datapoint from a fasta file. The structure and dms will be predicted if predict_structure and predict_dms are True."""
        sequences, references = Fasta.parse(fasta_file)
        return cls([DatapointFactory.from_fasta(sequence, reference, predict_structure, predict_dms)
                    for sequence, reference in tqdm_parser(zip(sequences, references), total=len(sequences), desc='Parsing fasta file', disable=not tqdm)], verbose=verbose)

    @classmethod
    def from_ct(cls, ct_files, predict_dms, tqdm=True, verbose=True):
        """Create a list of datapoint from a ct file. The dms will be predicted if predict_dms is True."""
        return cls([DatapointFactory.from_ct(ct_file, predict_dms) for ct_file in tqdm_parser(ct_files, total=len(ct_files), desc='Parsing ct files', disable=not tqdm)], verbose=verbose)

    @classmethod
    def from_dreem_output(cls, dreem_output_file, predict_structure, tqdm=True, verbose=True):
        """Create a list of datapoint from a dreem output file. The structure and dms will be predicted if predict_structure and predict_dms are True."""
        n_lines = len(list(DreemOutput.parse(dreem_output_file)))
        return cls([DatapointFactory.from_dreem_output(reference, sequence, mutation_rate, predict_structure)
            for reference, sequence, mutation_rate in tqdm_parser(DreemOutput.parse(dreem_output_file), total=n_lines, desc='Parsing dreem output file', disable=not tqdm)], verbose=verbose)

    @classmethod
    def from_json(cls, json_file, predict_structure=False, predict_dms=False, tqdm=True, verbose=True):
        """Create a list of datapoint from a json file."""
        data = json.load(open(json_file))
        return cls([DatapointFactory.from_json_line(reference, line, predict_structure, predict_dms) for reference, line in tqdm_parser(data.items(), total=len(data), desc='Parsing json file', disable=not tqdm)], verbose=verbose)


    def to_base_pairs_npy(self, path):
        """Writes the structure npy file from the list of datapoints.

        Args:
            path (str): path to the npy file

        Returns:
            np.array: structure matrix of pairs of 0-based indices.

        Examples:
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1,0,0,0,0,1])], verbose=False)
            >>> datapoints.to_base_pairs_npy('temp/base_pairs.npy')
            array([[[1, 2],
                    [3, 4]]], dtype=object)
        """

        arr = np.array([np.array(datapoint.paired_bases, dtype=np.uint8) for datapoint in self.datapoints], dtype=object)
        np.save(path, arr)
        return arr


    def to_dms_npy(self, path):
        """Writes the dms npy file from the list of datapoints.

        Args:
            path (str): path to the npy file

        Returns:
            np.array: dms matrix of floats arrays.

        Examples:
            >>> from numpy import array
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1,0,0,0,0,1])], verbose=False)
            >>> assert (datapoints.to_dms_npy('temp/dms.npy') == array([[1., 0., 0., 0., 0., 1.]], dtype=object)).all(), "The dms matrix is not correct."
        """
        
        arr = np.array([datapoint.embed_dms() for datapoint in self.datapoints], dtype=object)
        np.save(path, arr, allow_pickle=True)
        return arr


    def to_sequence_npy(self, path):
        """Writes the sequence npy file from the list of datapoints.

        Args:
            path (str): path to the npy file

        Returns:
            np.array: int-encoded sequence matrix

        Examples:
            >>> from numpy import array
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1,0,0,0,0,1]),\
                                               Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1,0,0,0,0,1])], verbose=False)
            >>> assert not (datapoints.to_sequence_npy('temp/sequence.npy') - array([[1, 1, 2, 2, 3, 3],[1, 1, 2, 2, 3, 3]], dtype=object)).any(), "The sequence matrix is not correct."
        """

        arr = np.array([datapoint.embed_sequence() for datapoint in self.datapoints], dtype=object)
        np.save(path, arr)
        return arr


    def to_reference_npy(self, path):
        """Writes the reference npy file from the list of datapoints.

        Args:
            path (str): path to the npy file

        Returns:
            np.array: reference matrix of int-encoded.

        Examples:
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1,0,0,0,0,1])], verbose=False)
            >>> datapoints.to_reference_npy('temp/reference.npy')
            array(['reference'], dtype=object)
        """

        arr = np.array([datapoint.reference for datapoint in self.datapoints], dtype=object)
        np.save(path, arr)
        return arr


    def to_json(self, path) -> None:
        """Write a list of datapoints to a json file."""
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'w') as f:
            f.write('{\n')
            for idx, datapoint in enumerate(self.datapoints): # write the datapoints one by one to avoid memory issues
                f.write(str(datapoint)+ ',\n' if idx != len(self.datapoints)-1 else str(datapoint)+'\n')
            f.write('}\n')

    def to_pandas(self, datapoints=None) -> pd.DataFrame:
        """Converts the list of datapoints into a pandas dataframe.

        Example:
        >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1,0,0,0,0,1])], verbose=False)
        >>> temp = datapoints.to_pandas()
        >>> temp['reference'].iloc[0]
        'reference'
        >>> temp['sequence'].iloc[0]
        'AACCGG'
        >>> temp['paired_bases'].iloc[0]
        ((1, 2), (3, 4))
        >>> temp['dms'].iloc[0]
        (1, 0, 0, 0, 0, 1)
        """
        if datapoints is None: datapoints = self.datapoints
        return pd.DataFrame([datapoint.to_flat_dict() for datapoint in datapoints])
    
    def from_pandas(self, df: pd.DataFrame) -> None:
        """Converts a pandas dataframe into a list of datapoints.

        Example:
            >>> df = pd.DataFrame([{'reference': 'reference', 'sequence': 'AACCGG', 'paired_bases': [[1, 2], [3, 4]], 'dms': [1.0, 2.0, 3.0]}])
            >>> ListofDatapoints(verbose=False).from_pandas(df)
            [Datapoint('reference', sequence='AACCGG', paired_bases=((1, 2), (3, 4)), dms=(1.0, 2.0, 3.0))]
        """
        return [Datapoint.from_flat_dict(datapoint_dict) for datapoint_dict in df.to_dict('records')]

    def filter_duplicates_and_unvalid(self, datapoints, verbose=True, min_AUROC:int=0.8):
        """Filters out duplicate sequences.
        Only keep the first occurence of a sequence if all the other structures are the same.

        Examples:
            >>> datapoints = ListofDatapoints([ Datapoint(reference='ref1', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1,0,0,0,0,1]),\
                                Datapoint(reference='ref2', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1,0,0,0,0,1]),\
                                Datapoint(reference='ref2', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1,0,0,0,0,1]),\
                                Datapoint(reference='ref4', sequence='AUGGC', paired_bases=[[1, 2]], dms=[0,0,0,0,1]),\
                                Datapoint(reference='ref5', sequence='AUGGC', paired_bases=[[0, 4], [2, 3]], dms=[0,0,0,0,0]),\
                                Datapoint(reference='ref6', sequence='not a regular sequence', paired_bases=[[0, 4]], dms=[0,0,0,0,0]) ])
            Over a total of 6 datapoints, there are:
                - 1 valid datapoints
                - 1 invalid datapoints (ex: sequence with non-regular characters)
                - 1 datapoints with the same reference
                - 1 duplicate sequences with the same structure / dms
                - 2 duplicate sequences with different structure / dms
                - 0 datapoints removed because of low AUROC (<0.8)
            >>> datapoints.datapoints
            [Datapoint('ref1', sequence='AACCGG', paired_bases=((1, 2), (3, 4)), dms=(1, 0, 0, 0, 0, 1))]
        """
        
        # Remove None datapoints        
        n_input_datapoints = len(datapoints)
        datapoints = list(filter(lambda x: x is not None, datapoints))
        n_unvalid_datapoints = n_input_datapoints - len(datapoints)

        # Convert to pandas
        df = self.to_pandas(datapoints)
        
        # Remove duplicate or conflicting datapoints and keep track of the number of datapoints removed
        def drop_duplicates(df:pd.DataFrame, **kwargs):
            len_df_before = len(df)
            df.drop_duplicates(**kwargs)
            return len_df_before - len(df)
        
        # Don't allow multiple sequences with the same reference
        n_same_ref_datapoints = drop_duplicates(df, subset=['reference'], inplace=True)

        # Keep only one datapoint per sequence and structure
        if 'paired_bases' in df.columns:
            n_duplicates_datapoints = drop_duplicates(df, subset=['sequence','paired_bases'], inplace=True, ignore_index=True)

        # Keep only one datapoint per sequence and dms
        if 'dms' in df.columns:
            if not 'paired_bases' in df.columns: n_duplicates_datapoints = 0
            n_duplicates_datapoints += drop_duplicates(df, subset=['sequence','dms'], inplace=True, ignore_index=True)
        
        # If there are multiple structures / dms with the same structure, keep none
        n_same_seq_datapoints = drop_duplicates(df, subset=['sequence'], inplace=True, ignore_index=True, keep='first' if not ('paired_bases' in df.columns or 'dms' in df.columns) else False)


        ## Filter out references with low AUROC 
        if 'paired_bases' in df.columns and 'dms' in df.columns:
            def calculate_auroc(row):
                dms = np.array(row['dms'])
                isUnpaired = np.ones_like(dms)
                isUnpaired[np.array(row['paired_bases']).flatten()] = 0
                return roc_auc_score(isUnpaired[dms!=UKN], dms[dms!=UKN])

            # Create a boolean mask for rows with auroc score greater than or equal to a threshold
            mask_high_AUROC = df.apply(lambda row: calculate_auroc(row) >= min_AUROC, axis=1)
            df = df[mask_high_AUROC]

        # Convert back to list of datapoints
        datapoints = self.from_pandas(df)

        # Write report
        report = f"""Over a total of {n_input_datapoints} datapoints, there are:
    - {len(datapoints)} valid datapoints
    - {n_unvalid_datapoints} invalid datapoints (ex: sequence with non-regular characters)
    - {n_same_ref_datapoints} datapoints with the same reference"""
        if 'paired_bases' in df.columns or 'dms' in df.columns:
            report += f"""
    - {n_duplicates_datapoints} duplicate sequences with the same structure / dms
    - {n_same_seq_datapoints} duplicate sequences with different structure / dms"""
        else:
            report += f"""
    - {n_same_seq_datapoints} duplicate sequences"""
        if 'paired_bases' in df.columns and 'dms' in df.columns:
            report += f"""
    - {np.sum(~mask_high_AUROC)} datapoints removed because of low AUROC (<{min_AUROC})"""
    
        if verbose: print(report)
            
        return datapoints, report
    
            
def base_pairs_into_matrix(base_pairs):
    """Turns a list of base pairs into a matrix of 1s and 0s.

    Args:
        base_pairs (list): list of base pairs, where each base pair is a tuple of 0-based two ints.

    Returns:
        matrix (np.array): matrix of 1s and 0s.

    Example:
        >>> from numpy import array, int64
        >>> base_pairs = [(0, 1), (2, 3), (4, 5)]
        >>> assert (base_pairs_into_matrix(base_pairs) ==  array([[0., 1., 0., 0., 0., 0.], \
                                                                [1., 0., 0., 0., 0., 0.], \
                                                                [0., 0., 0., 1., 0., 0.], \
                                                                [0., 0., 1., 0., 0., 0.], \
                                                                [0., 0., 0., 0., 0., 1.], \
                                                                [0., 0., 0., 0., 1., 0.]], dtype=int64)).all()

    """
    max_index = max([max(base_pair) for base_pair in base_pairs])
    arr = np.zeros((max_index + 1, max_index + 1))
    for base_pair in base_pairs:
        arr[base_pair[0], base_pair[1]] = 1
        arr[base_pair[1], base_pair[0]] = 1
    return arr